unit FormatCadFiles;

interface

uses
  Windows, Classes, Graphics,
  FlexBase, FlexUtils, FlexPath, FlexControls, FlexProps, FlexFileFormats,
  CADIntf, sgConsts, sgLines;

const
  CADImportExtensions =
    'dwg,dxf,plt,hgl,hg,hpg,plo,hp,hp1,hp2,hpgl,hpgl2,gl2,prn,spl,rtl,cgm,svg,svgz';

resourcestring
  sDxfCadFileDescription = 'CAD library - %s File';
  sDxfCadNeedFilename    = 'Can''t load file from steam. Need Filename.';

type
  TFlexCADFormat = class(TFlexFileFormat)
  private
    FFlex: TFlexPanel;
    FCAD: THandle;
    FPenWidth: integer;
    FLayerColors: array of integer;
    FLines: TsgLines;
    function LoadCurveGlobal(SPoints: TList; IsSolid: Boolean): TFlexCurve;
    function GetPoint(const P: TFPoint): TPoint;
  protected
    FX: Integer;
    FY: Integer;
    FScale: double;
    procedure RegisterSupportedExtensions; override;
    procedure LoadEntity(Data: PcadData);
    function LoadLine(Data: PcadData): TFlexCurve;
    function LoadPoly(Data: PcadData): TFlexCurve;
    function LoadSpline(Data: PcadData): TFlexCurve;
    function Load3DFace(Data: PcadData): TFlexCurve;
    function LoadArc(Data: PcadData): TFlexEllipse;
    function LoadSolid(Data: PcadData): TFlexCurve;
    function LoadHatch(Data: PcadData): TFlexCurve;
    function LoadText(Data: PcadData): TFlexText;
    function LoadImage(Data: PcadData): TFlexPicture;
  public
    constructor Create(AOwner: TObject); override;
    destructor Destroy; override;
    procedure ImportFromStream(AStream: TStream; AFlexPanel: TFlexPanel;
      const Extension: TFlexFileExtension; const AFileName: string); override;
    function ImportFromFile(AFlex: TFlexPanel; const AFileName: string;
      AllLayers: boolean; AScale: double = 0; ADocName: string = '';
      ANewDocument: boolean = True): boolean;
    function ImportFromCad(AFlex: TFlexPanel; ACad: THandle; AllLayers: boolean;
      AScale: double; ADocName: string = '';
      ANewDocument: boolean = True): boolean;
    property Flex: TFlexPanel read FFlex;
    property CadHandle: THandle read FCAD;
  end;

implementation

uses
  SysUtils, Math;

{ GetSpline

  Gets a spline.  }
function GetSpline(ControlPoints, FitPoints, Knots: TList; ChordLen: Single): TList;

  function Normalize(N, I: Integer; Param: Double): Double;
  var
    V1, D1, V2, D2: Double;
  begin
    if N = 0 then
    begin
      if (Single(Knots[I]) <= Param) and (Param < Single(Knots[I + 1])) then
        Result := 1
      else
        Result := 0;
    end
    else
    begin
      D1 := Single(Knots[I + N]) - Single(Knots[I]);
      V1 := (Param - Single(Knots[I])) * Normalize(N - 1, I, Param);
      if D1 = 0 then
        V1 := 0
      else
        V1 := V1 / D1;
      D2 := Single(Knots[I + N + 1]) - Single(Knots[I + 1]);
      V2 := (Single(Knots[I + N + 1]) - Param) * Normalize(N - 1, I + 1, Param);
      if D2 = 0 then
        V2 := 0
      else
        V2 := V2 / D2;
      Result := V1 + V2;
    end;
  end;

  function GetNURBS(Index: Integer; Param: Double): TFPoint;
  var
    I: Integer;
    Ni: Double;
  begin
    Result.X := 0;
    Result.Y := 0;
    Result.Z := 0;
    for I := Index - 3 to Index do
    begin
      Ni := Normalize(3, I, Param);
      Result.X := Result.X + PFPoint(ControlPoints[I]).X * Ni;
      Result.Y := Result.Y + PFPoint(ControlPoints[I]).Y * Ni;
      Result.Z := Result.Z + PFPoint(ControlPoints[I]).Z * Ni;
    end;
  end;

  procedure AddPt(const Pt: TFPoint);
  var
    C: Integer;
    P: PInteger;
  begin
    C := Result.Count;
    Result.Count := Result.Count + cnstFPointSize;
    P := PInteger(Result.List);
    Inc(P,C);
    Move(Pt, P^, SizeOf(TFPoint));
  end;

var
  I: Integer;
  P, Prev: TFPoint;
  T: Double;
begin
  Result := TList.Create;
  P := PFPoint(ControlPoints[0])^;
  AddPt(P);
  Prev := P;
  for I := 3 to ControlPoints.Count - 1 do
  begin
    T := Single(Knots[I]);
    while  T < Single(Knots[I + 1]) do
    begin
      P := GetNURBS(I, T);
      T := T + 0.1; // the smaller value (for instance 0.01) the higher quality
      if (ChordLen > 0) and (abs(P.X - Prev.X) < ChordLen) and (abs(P.Y - Prev.Y) < ChordLen)
      then Continue;
      AddPt(P);
      Prev := P;
    end;
  end;
  P := PFPoint(ControlPoints[ControlPoints.Count - 1])^;
  AddPt(P);
end;

procedure CADEnumProc(Data: PcadData; var Param); stdcall;
begin
  TFlexCADFormat(Param).LoadEntity(Data);
end;

procedure CADError;
var Buf: array[Byte] of AnsiChar;
begin
  CADGetLastError(Buf);
  raise Exception.Create(String(Buf));
end;

// TFlexCADFormat /////////////////////////////////////////////////////////////

constructor TFlexCADFormat.Create(AOwner: TObject); 
begin
 inherited;
 FStreamSupport := False;
 FLines := TsgLines.Create;
 FPenWidth := PixelScaleFactor div 100;
 FScale := 100 * PixelScaleFactor;
end;

destructor TFlexCADFormat.Destroy;
begin
 inherited;
 FLines.Free;
end;

procedure TFlexCADFormat.RegisterSupportedExtensions;
var
  ExtList: TStringList;
  i: integer;
begin
  ExtList := TStringList.Create;
  try
    ExtList.CommaText := CADImportExtensions;
    for i:=0 to ExtList.Count-1 do
      RegisterExtension(
        ExtList[i],
        Format(sDxfCadFileDescription, [UpperCase(ExtList[i])]),
        [skImport]
      );
  finally
    ExtList.Free;
  end;
end;

procedure TFlexCADFormat.ImportFromStream(AStream: TStream;
  AFlexPanel: TFlexPanel; const Extension: TFlexFileExtension;
  const AFileName: string);
begin
  if AFileName = '' then
    raise Exception.Create(sDxfCadNeedFilename);
  ImportFromFile(AFlexPanel, AFileName, True);
end;

function TFlexCADFormat.ImportFromFile(AFlex: TFlexPanel; const AFileName: string;
  AllLayers: boolean; AScale: double = 0; ADocName: string = '';
  ANewDocument: boolean = True): boolean;
var
  CADHandle: THandle;
  L, T, R, B, W, H: Double;
begin
  CADHandle := CADCreate(0, PAnsiChar(AnsiString(AFileName)));    // Open CADHandle file
  if CADHandle = 0 then CADError;
  try
    if AScale = 0 then begin
      if CADGetBox(CADHandle, L, R, T, B) = 0 then CADError;
      W := R - L;
      if W = 0 then W := 1;
      H := T - B;
      if H = 0 then H := 1;
      W := AFlex.DocWidth * 100 / (W * PixelScaleFactor);
      H := AFlex.DocHeight * 100 / (H * PixelScaleFactor);
      if W > H
        then AScale := H
        else AScale := W;
    end;
    Result := ImportFromCad(AFlex, CADHandle, AllLayers, AScale);
  finally
    CADClose(CADHandle);
  end;
end;

function TFlexCADFormat.ImportFromCad(AFlex: TFlexPanel; ACad: THandle;
  AllLayers: boolean; AScale: double; ADocName: string = '';
  ANewDocument: boolean = True): boolean;
var i: integer;
    R: TFRect;
    Count: integer;
    Data: TcadData;
    Layer: TFlexLayer;
    LScale: double;
begin
 Result := false;
 if Assigned(FFlex) or (FCad <> 0) or
    not Assigned(AFlex) or (ACad = 0) then exit;
 try
  FFlex := AFlex;
  if ANewDocument then
    FFlex.NewDocument;
  FCAD := ACad;
  // Set name
  if ADocName <> '' then
    FFlex.Schemes.Name := ADocName;
  // Set document size
  if CADLTScale(CADHandle, LScale) = 0 then
    exit;
  if CADGetBox(FCad, R.Left, R.Right, R.Top, R.Bottom) = 0 then
    exit;
  FLines.Scale := LScale;
  FScale := Round(AScale * PixelScaleFactor);
  FFlex.DocWidth := Round((R.Right - R.Left) * FScale / 100);
  FFlex.DocHeight := Round((R.Top - R.Bottom) * FScale / 100);
  FX := - Round(R.Left * FScale / 100); // see GetPoint
  FY := Round(R.Top * FScale / 100);
  // Process layers
  Count := CADLayerCount(CADHandle);
  if ANewDocument and (Count > 0) then
    FFlex.Layers.Clear;
  SetLength(FLayerColors, Count);
  for i:=0 to Count-1 do begin
    CADLayer(FCad, i, @Data);
    FLayerColors[i] := Data.Color;    
    Layer := FFlex.Layers.ByName[String(Data.Text)];
    if not Assigned(Layer) then begin
      Layer := FFlex.Layers.New;
      Layer.Name := String(Data.Text);
    end;
    if Data.Flags and 1 <> 0 then
      // Invisible
      Layer.Visible := False;
  end;
  FLines.Scale := FScale;
  CADEnum(FCad, integer(AllLayers), @CADEnumProc, Self);
 finally
  FFlex := Nil;
  FCad := 0;
 end;
 Result := true;
end;

procedure TFlexCADFormat.LoadEntity(Data: PcadData);
type
  PSingle = ^Single;
var
  i: Integer;
  p: PSingle;
  Control: TFlexControl;
  PenProp: TPenProp;
  BrushProp: TBrushProp;
  Layer: TFlexLayer;
  ControlColor: TColor;
begin
  // Fill lines
  FLines.List.Clear;
  FLines.Scale := Data.Rotation * FScale;
  p := PSingle(Data.Ticks);
  for i:=0 to Data.TickCount-1 do begin
    FLines.AddTick(p^);
    Inc(p);
  end;
  // Check entity
  Control := Nil;
  case Data.Tag of
    DXF_LINE:
      Control := LoadLine(Data);
    DXF_POLYLINE..DXF_LWPOLYLINE:
      Control := LoadPoly(Data);
    DXF_CIRCLE,DXF_ARC,DXF_ELLIPSE:
      Control := LoadArc(Data);
    DXF_SPLINE:
      Control := LoadSpline(Data);
    DXF_SOLID:
      Control := LoadSolid(Data);
    DXF_HATCH:
      Control := LoadHatch(Data);
    DXF_TEXT..DXF_ATTDEF, DXF_ATTRIB:
      Control := LoadText(Data);
    DXF_3DFACE:
      Control := Load3DFace(Data);
    DXF_IMAGE_ENT:
      Control := LoadImage(Data);
    DXF_BEGIN_INSERT:
      ; // FControlParent := TFlexGroup.Create(FFlex, FControlParent, FFlex.ActiveLayer);
    DXF_END_INSERT:
      ; // FControlParent := FControlParent.Parent;

{
    DXF_POINT:				DrawPoint;

    DXF_BEGIN_INSERT:                   IsInsideInsert := True;
    DXF_END_INSERT:                     IsInsideInsert := False;
    DXF_BEGIN_VIEWPORT:                 BeginViewport;
    DXF_END_VIEWPORT:                   EndViewport;  }
  end;
  if Assigned(Control) then begin
    Layer := TFlexLayer(FFlex.Layers.FindByName(String(Data.Layer)));
    if Assigned(Layer) then
      Control.Layer := Layer;
    if (Data.Color = clNone) or (Data.Color = clDefault) then
      ControlColor := clBlack
    else
      ControlColor := Data.Color;
    PenProp := TPenProp(Control.Props['Pen']);
    if Control is TFlexText then
      TFlexText(Control).FontProp.Color := ControlColor
    else
    if Assigned(PenProp) then
      if Data.Tag = DXF_HATCH then
        PenProp.Style := psClear
      else begin
        PenProp.Color := ControlColor;
        i := Round(Data.Thickness * FScale / 100);
        if i <= 0 then
          // Minimal thickness
          PenProp.Width := 1
        else
          PenProp.Width := i;
      end;
    if Data.Tag = DXF_HATCH then begin
      BrushProp := TBrushProp(Control.Props['Brush']);
      if Assigned(BrushProp) then
        BrushProp.Color := ControlColor;
    end;
  end;
end;

function TFlexCADFormat.LoadLine(Data: PcadData): TFlexCurve;
var
  I: Integer;
  SingPoints, DottedSingPoints: TList;
  PSinglePt: PFPoint;
begin
  SingPoints := TList.Create;
  DottedSingPoints := TList.Create;
  New(PSinglePt);
  PSinglePt^.X := Data.Point.X;
  PSinglePt^.Y := Data.Point.Y;
  PSinglePt^.Z := Data.Point.Z;
  SingPoints.Add(PSinglePt);
  New(PSinglePt);
  PSinglePt^.X := Data.Point1.X;
  PSinglePt^.Y := Data.Point1.Y;
  PSinglePt^.Z := Data.Point1.Z;
  SingPoints.Add(PSinglePt);

  if FLines.IsSolid then
    Result := LoadCurveGlobal(SingPoints, FLines.IsSolid)
  else
  begin
    FLines.Line(TFPoint(SingPoints[0]^), TFPoint(SingPoints[1]^), DottedSingPoints);
    Result := LoadCurveGlobal(DottedSingPoints, FLines.IsSolid);
  end;

  // Clearing memory
  for I := 0 to DottedSingPoints.Count - 1 do
    Dispose(DottedSingPoints[I]);
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  DottedSingPoints.Free;
  SingPoints.Free;
end;

function TFlexCADFormat.LoadPoly(Data: PcadData): TFlexCurve;
var
  I: Integer;
  P: PFPoint;
  PSinglePt: PFPoint;
  SingPoints, DottedSingPoints: TList;
begin
  SingPoints := TList.Create;
  DottedSingPoints := TList.Create;

  P := Data.Points;
  for I := 0 to Data.Count - 1 do
  begin
    New(PSinglePt);
    PSinglePt^ := P^;
    SingPoints.Add(PSinglePt);
    Inc(P);
  end;

  if FLines.IsSolid then
    Result := LoadCurveGlobal(SingPoints, FLines.IsSolid)
  else begin
    FLines.Poly(SingPoints, DottedSingPoints, False);
    Result := LoadCurveGlobal(DottedSingPoints, FLines.IsSolid);
  end;

  // Clearing memory
  for I := 0 to DottedSingPoints.Count - 1 do
    Dispose(DottedSingPoints[I]);
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  DottedSingPoints.Free;
  SingPoints.Free;
end;

function TFlexCADFormat.LoadSpline(Data: PcadData): TFlexCurve;
var
  I: Integer;
  S: Single;
  ControlPoints, FitPoints, Knots, vList: TList;
  P, vP: PFPoint;
  SingPoints, DottedSingPoints: TList;
  PSinglePt: PFPoint;
begin
  SingPoints := TList.Create;
  DottedSingPoints := TList.Create;
  ControlPoints := TList.Create;
  FitPoints := TList.Create;
  Knots := TList.Create;
  try
    P := Data.Points;
    for I := 0 to Data.Count - 1 do begin
      New(vP);
      vP^ := P^;
      ControlPoints.Add(vP);
      Inc(P);
    end;
    for I := 0 to PInteger(P)^ - 1 do begin
      Inc(P);
      New(vP);
      vP^ := P^;
      FitPoints.Add(vP);
    end;
    Inc(P);
    for I := 0 to PInteger(P)^ - 1 do begin
      Inc(P);
      S := P.X;
      Knots.Add(Pointer(S));
    end;

    vList := Nil;
    try
      if FitPoints.Count > ControlPoints.Count then begin
        vList := TList.Create;
        for I := 0 to FitPoints.Count - 1 do begin
          New(PSinglePt);
          PSinglePt^.X := PFPoint(FitPoints[I]).X;
          PSinglePt^.Y := PFPoint(FitPoints[I]).Y;
          PSinglePt^.Z := PFPoint(FitPoints[I]).Z;
          SingPoints.Add(PSinglePt);
        end;
        Data.Points := PFPoint(vList.List);
        Data.Count := FitPoints.Count;
      end else begin
        vList := GetSpline(ControlPoints, FitPoints, Knots, 0);
        Data.Points := PFPoint(vList.List);
        Data.Count := vList.Count div cnstFPointSize;
        P := Data.Points;
        for I := 0 to Data.Count-1 do begin
          New(PSinglePt);
          PSinglePt^.X := P^.X;
          PSinglePt^.Y := P^.Y;
          PSinglePt^.Z := P^.Z;
          SingPoints.Add(PSinglePt);
          Inc(P);
        end;
      end;
    finally
      vList.Free;
    end;

    if FLines.IsSolid then
      Result := LoadCurveGlobal(SingPoints, FLines.IsSolid)
    else begin
      FLines.Curve(SingPoints, DottedSingPoints, False);
      Result := LoadCurveGlobal(DottedSingPoints, False);
    end;

  finally
    // Clearing memory
    for I := 0 to DottedSingPoints.Count - 1 do Dispose(DottedSingPoints[I]);
    for I := 0 to ControlPoints.Count - 1 do Dispose(ControlPoints[I]);
    for I := 0 to SingPoints.Count - 1 do Dispose(SingPoints[I]);
    for I := 0 to FitPoints.Count - 1 do Dispose(FitPoints[I]);
    DottedSingPoints.Free;
    ControlPoints.Free;
    SingPoints.Free;
    FitPoints.Free;
    Knots.Free;
  end;
end;

function TFlexCADFormat.Load3DFace(Data: PcadData): TFlexCurve;

  procedure Edge(const P1, P2: TFPoint);
  begin
    Result.AddPoint(GetPoint(P1));
    Result.AddPoint(GetPoint(P2));
    Result.EndFigure(True);
  end;

begin
  Result := TFlexCurve.Create(FFlex, FFlex.ActiveScheme, FFlex.ActiveLayer);
  Result.BeginUpdate;
  try
    while Result.PointCount > 0 do Result.DeletePoint(0);
    if Data.Flags and 1 = 0 then Edge(Data.Point, Data.Point1);
    if Data.Flags and 2 = 0 then Edge(Data.Point1, Data.Point2);
    if Data.Flags and 4 = 0 then Edge(Data.Point2, Data.Point3);
    if Data.Flags and 8 = 0 then Edge(Data.Point3, Data.Point);
  finally
    Result.EndUpdate;
  end;
end;

function TFlexCADFormat.LoadArc(Data: PcadData): TFlexEllipse;
var
  P,P1,P2,P3: TPoint;
  R: TRect;

  function CalcAngle(X, Y: double; const Coeff: double): integer;
  var Angle: double;
  begin
    Y := Y * Coeff;
    if X = 0 then begin
     if Y > 0
      then Result := 90 * PixelScaleFactor
      else Result := 270 * PixelScaleFactor;
    end else begin
     Angle := 180.0 * ArcTan2(Y, X) / pi;
     if Angle < 0 then Angle := 360.0 + Angle;
     Result := Round(Angle * PixelScaleFactor);
    end;
  end;

  procedure SetAnglesByPoints(const P0, P1: TPoint; const DocRect: TRect);
  var HR, WR: double;
      Coeff: double;
  begin
    WR := Result.WidthProp.Value / 2;
    HR := Result.HeightProp.Value / 2;
    if HR = 0
     then Coeff := 0
     else Coeff := WR / HR;
    with DocRect do begin
      Result.BeginAngleProp.Value := CalcAngle(P0.X - Left - WR, HR - P0.Y + Top, Coeff);
      Result.EndAngleProp.Value   := CalcAngle(P1.X - Left - WR, HR - P1.Y + Top, Coeff);
    end;
  end;

begin
  P := GetPoint(Data.Point);
  P1 := GetPoint(Data.Point1);
  P2 := GetPoint(Data.Point2);
  P3 := GetPoint(Data.Point3);
  Result := TFlexEllipse.Create(FFlex, FFlex.ActiveScheme, FFlex.ActiveLayer);
  R := Rect(P.X, P.Y, P1.X, P1.Y);
  Result.DocRect := R;
  if (P2.X <> P3.X) or (P2.Y <> P3.Y) then
    SetAnglesByPoints(P2, P3, R);
end;

function TFlexCADFormat.LoadSolid(Data: PcadData): TFlexCurve;
var
  I: Integer;
  PSinglePt: PFPoint;
  SingPoints: TList;
begin
  SingPoints := TList.Create;
  New(PSinglePt);
  PSinglePt^ := Data.Point;
  SingPoints.Add(PSinglePt);
  New(PSinglePt);
  PSinglePt^ := Data.Point1;
  SingPoints.Add(PSinglePt);
  New(PSinglePt);
  PSinglePt^ := Data.Point2;
  SingPoints.Add(PSinglePt);
  New(PSinglePt);
  PSinglePt^ := Data.Point3;
  SingPoints.Add(PSinglePt);

  Result := LoadCurveGlobal(SingPoints, True);

  // Clearing memory
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  SingPoints.Free;
end;

{ DrawHatch

  Draws a hatch.
  EData.Count - number of polyline vertices
  EData.Points - pointer to point array }
function TFlexCADFormat.LoadHatch(Data: PcadData): TFlexCurve;
var
  I, Index: Integer;
  P: PFPoint;
begin
  Result := TFlexCurve.Create(FFlex, FFlex.ActiveScheme, FFlex.ActiveLayer);
  Result.BeginUpdate;
  try
    while Result.PointCount > 0 do Result.DeletePoint(0);
    if Data.Flags and 16 <> 0 then begin
      // hatch is SOLID
      P := Data.Points;
      Index := 0;
      for I := 0 to Data.Count - 1 do begin
        if Index = I then begin
          Inc(Index, PInteger(P)^ + 1);
          if I <> 0 then
            Result.EndFigure(True);
        end else
          Result.AddPoint(GetPoint(P^));
        Inc(P);
      end;
      Result.EndFigure(True);
    end;
  finally
    Result.EndUpdate;
  end;

{
  if Data.Flags and 16 <> 0 then begin        // hatch is SOLID
    vPoly := TList.Create;
    vMainRGN := CreateRectRgn(0, 0, 0, 0);// FCanvas.ClipRect.Right, FCanvas.ClipRect.Bottom);
    SaveIndex := SaveDC(PCanvas.Handle);
    try
      P := EData.Points;
      Index := 0;
      for I := 0 to EData.Count do begin
        if Index = I then begin
          Inc(Index, PInteger(P)^ + 1);
          if I <> 0 then begin
            vRGN := CreatePolygonRgn(vPoly.List^, vPoly.Count shr 1, ALTERNATE);
            CombineRgn(vMainRGN, vRGN, vMainRGN,  RGN_XOR);
            DeleteObject(vRGN);
            vPoly.Clear;
          end
        end else begin
          P1 := GetPoint(P^);
          vPoly.Add(Pointer(P1.X));
          vPoly.Add(Pointer(P1.Y));
        end;
        Inc(P);
      end;
      I := GetRegionData(vMainRgn,0,nil);
      PCanvas.Brush.Style := bsSolid;
      PCanvas.Brush.Color := EData.Color;
      FillRgn(PCanvas.Handle, vMainRGN, PCanvas.Brush.Handle);
    finally
      RestoreDC(PCanvas.Handle, SaveIndex);
      DeleteObject(vMainRGN);
      vPoly.Free;
    end;
    Exit;
  end;
  SingPoints := TList.Create;
  P := EData.DashDots;
  for I := 0 to EData.DashDotsCount - 1 do
  begin
    New(PSinglePt);
    PSinglePt^.X := P^.X;
    PSinglePt^.Y := P^.Y;
    PSinglePt^.Z := P^.Z;
    SingPoints.Add(PSinglePt);
    Inc(P);
  end;
  DrawGlobal(SingPoints, False);
  // Clearing memory
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  SingPoints.Free;             }
end;

function TFlexCADFormat.LoadText(Data: PcadData): TFlexText;
var
  P: TPoint;
  vFontName: string;
  I, Len: Integer;
  Bmp: TBitmap;
  tm: TTextMetric;
  ti: TTranslateInfo;
begin
  Result := TFlexText.Create(FFlex, FFlex.ActiveScheme, FFlex.ActiveLayer);
  if Data.HAlign in [1,2,4] then
    P := GetPoint(Data.Point1)
  else
    P := GetPoint(Data.Point);

  vFontName := string(Data.FontName);
  Len := Length(vFontName);
  if Len = 0 then
    vFontName := 'Arial'
  else begin // setting real TrueType font
    I := Pos('.', vFontName);
    if I > 0 then
      SetLength(vFontName, I - 1);
  end;
  with Result.FontProp do begin
    Name := vFontName;
    if Data.FHeight = 0 then
      Height := PixelScaleFactor
    else
      Height := Round(fWinTextHeightFactor * Data.FHeight * FScale / 100);
  end;
  Result.TextProp.Text := String(Data.Text);

  Bmp := TBitmap.Create;
  try
    Result.FontProp.Setup(Bmp.Canvas, 100 * PixelScaleFactor);
    GetTextMetrics(Bmp.Canvas.Handle, tm);
    dec(P.Y, tm.tmAscent);
  finally
    Bmp.Free;
  end;

  Result.DocRect := Rect(P.X, P.Y, P.X, P.Y);
  Result.AutoSizeProp.Value := True;

  Result.IsSelected := True;
  with Result.DocRect do
    ti.Center := Point(Left, Bottom);
  ti.Rotate := Round(Data.Rotation);
  if ti.Rotate < 0 then
    ti.Rotate := 360 - Abs(ti.Rotate) mod 360;
  ti.Mirror := False;
  FFlex.Translate(ti);
  Result.IsSelected := False;
end;

function TFlexCADFormat.LoadImage(Data: PcadData): TFlexPicture;
var
  MS: TMemoryStream;
  R: TRect;
begin
  Result := TFlexPicture.Create(FFlex, FFlex.ActiveScheme, FFlex.ActiveLayer);
  R.TopLeft := GetPoint(Data.Point1);
  R.BottomRight := GetPoint(Data.Point2);
  Result.DocRect := R;
  MS := TMemoryStream.Create;
  try
    MS.Write(Data.Ticks^, Data.Handle);
    MS.Position := 0;
    Result.PictureProp.LoadFromStream(MS);
  finally
    MS.Free;
  end;
end;

{ GetPoint

  Converts CADHandle units to pixels.}
function TFlexCADFormat.GetPoint(const P: TFPoint): TPoint;
begin
  Result.X := (FX + Round(P.X * FScale / 100));
  Result.Y := (FY - Round(P.Y * FScale / 100));
end;

{ DrawGlobal

  Draws a (poly)line in global coordinates. }
function TFlexCADFormat.LoadCurveGlobal(SPoints: TList; IsSolid: Boolean): TFlexCurve;
var
  I, N: Integer;
  Prev, PA, PB: TPoint;
  PDXF: TFPoint;
begin
  if SPoints.Count = 0 then
    Result := Nil // Error
  else begin
    Result := TFlexCurve.Create(FFlex, FFlex.ActiveScheme, FFlex.ActiveLayer);
    Result.BeginUpdate;
    try
      while Result.PointCount > 0 do Result.DeletePoint(0);
      N := 2;
      if IsSolid then
        N := 1;
      I := 0;
      while I < SPoints.Count -1 do begin
        PDXF.X := PFPoint(SPoints.Items[I])^.X;
        PDXF.Y := PFPoint(SPoints.Items[I])^.Y;
        PDXF.Z := PFPoint(SPoints.Items[I])^.Z;// for future version
        PA := GetPoint(PDXF);
        PDXF.X := PFPoint(SPoints.Items[I+1])^.X;
        PDXF.Y := PFPoint(SPoints.Items[I+1])^.Y;
        PDXF.Z := PFPoint(SPoints.Items[I+1])^.Z;// for future version
        PB := GetPoint(PDXF);
        if (PA.X = Prev.X) and (PA.Y = Prev.Y) then
          Result.AddPoint(PB)
        else begin
          Result.EndFigure(False);
          Result.AddPoint(PA);
          Result.AddPoint(PB);
        end;
        Inc(I, N);
        Prev := PB;
      end;
      if Result.PointTypes[Result.PointCount-1] = ptNode then
        Result.EndFigure(False);
    finally
      Result.EndUpdate;
    end;
  end;
end;

initialization
  RegisteredFlexFileFormats.RegisterFormat(TFlexCADFormat);

end.
